/**
 * \file fire.c
 * \brief Fire Effect Demo
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000-2002
 * by Thomas Nussbaumer & Lionel Debroux, with two modifications:
 * - this version uses a LFSR (pseudo-RNG) for emulating the flame decay
 *   effect, instead of a larger and slower RNG;
 * - this version provides a way to adjust contrast during execution, in
 *   order to make it easier to see moving objects on that slow screen
 *   of terrible quality.
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>
#include <libndls.h>
#include <stdint.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


//! Generator for 32-bit numbers.
#define LFSR_XOR_MASK (0xA3000000)

#define BITS_PER_PIXEL  (4)
#define BITS_PER_BYTE   (8)
#define SCREEN_BYTES_PER_LINE (SCREEN_WIDTH * BITS_PER_PIXEL / BITS_PER_BYTE)

#define RESERVED_AREA_AT_TOP (80)

//! Used as offscreen buffer to generate image
uint8_t * backbuffer[2] = {NULL, NULL};
//! Holds the LFSR value
uint32_t seq;


/**
 * Allocates memory for backup buffer and clears contents
 *
 * NOTE: the backbuffers are allocated in one piece, split afterwards so as
 *       to save on 1) system handles and 2) code to free each individual
 *       buffer.
 */
static inline short InitBackBuffer(void) {
   if (!(backbuffer[0] = malloc(2 * SCREEN_BYTES_SIZE))) return 0;

    backbuffer[1] = backbuffer[0] + SCREEN_BYTES_SIZE;
    memset(backbuffer[0], 0xFF, 2 * SCREEN_BYTES_SIZE);
    return 1;
}


/**
 * Copies offscreen image to visible bitplanes
 */
static inline void BlitScreen(void) {
    memcpy(SCREEN_BASE_ADDRESS + RESERVED_AREA_AT_TOP * SCREEN_BYTES_PER_LINE,
           backbuffer[1] + RESERVED_AREA_AT_TOP * SCREEN_BYTES_PER_LINE,
           SCREEN_BYTES_SIZE - RESERVED_AREA_AT_TOP * SCREEN_BYTES_PER_LINE);

    memcpy(backbuffer[0], backbuffer[1], SCREEN_BYTES_SIZE);
}


/**
 * Fill up bottom lines in first backbuffer
 */
static inline void FillBottomLines(void) {
    uint32_t x;
    uint32_t * ptr;

    // Fill last 3 lines with black.
    memset(backbuffer[0] + SCREEN_BYTES_SIZE - 3 * SCREEN_BYTES_PER_LINE, 0x00, 3 * SCREEN_BYTES_PER_LINE);

    // Fill 4th line from bottom with random garbage.
    ptr = (uint32_t *)(backbuffer[0] + SCREEN_BYTES_SIZE - 4 * SCREEN_BYTES_PER_LINE);
    for (x = 0; x < SCREEN_BYTES_PER_LINE / 4; x++) {
        *ptr++ = seq;
        if (seq & 1) {
            seq = (seq >> 1) ^ LFSR_XOR_MASK;
        }
        else {
            seq = seq >> 1;
        }
    }
}


/**
 * Copies first backbuffer into second backbuffer with one line offset so
 * it looks like as scrolling upwards and modify some pixels randomly afterwards
 */
static inline void ScrollUp(void) {
    uint16_t *ptr;

    memcpy(backbuffer[1], backbuffer[0] + SCREEN_BYTES_PER_LINE, SCREEN_BYTES_SIZE - SCREEN_BYTES_PER_LINE);

    // C89_92(2000,2560) => LCD_HEIGHT*2*30/3
    // C89_92(2910,3750) => (LCD_HEIGHT-3)*30
    for (ptr = (uint16_t *)(backbuffer[1] + RESERVED_AREA_AT_TOP * SCREEN_BYTES_PER_LINE);
         ptr < (uint16_t *)(backbuffer[1] + SCREEN_BYTES_SIZE - 7 * SCREEN_BYTES_PER_LINE);
         ptr++)
    {
        if (seq & 1) {
            seq = (seq >> 1) ^ LFSR_XOR_MASK;
        }
        else {
            seq = seq >> 1;
        }

        if (seq <= 0x1FFFFFFF) *ptr |= (uint16_t)seq;

    }
}


/**
 * Main routine: where all the fun starts ...
 */
int main(void) {
    uint32_t initial_contrast = *((uint32_t *)0x900F0020);

    show_msgbox(
        "fire - Fire effect demo",
        "------------------------\n"
        "    Fire effect demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "BKSPC - Reset the screen\n"
        "RIGHT - Increase contrast\n"
        "LEFT  - Decrease contrast\n"
        "ESC   - Exit"
    );

    WAIT("0x1FFFFF", "wait");

    // Initialize the random number generator
    seq = 1234567891;

begin:

    // Try to allocate memory for offscreen buffers
    if (InitBackBuffer()) {
        // Clear both planes
        memset(SCREEN_BASE_ADDRESS, 0xFF, SCREEN_BYTES_SIZE);

        // Main loop
        while (!isKeyPressed(KEY_NSPIRE_ESC)) {
            FillBottomLines();
            ScrollUp();
            BlitScreen();

            WAIT("0x1FFFFF", "loop_slowdown");

            if (isKeyPressed(KEY_NSPIRE_CLEAR)) {
                // Free the offscreen buffers.
                // (else we will have memory leaks, 15360 bytes each time).
                free(backbuffer[0]);
                goto begin;
            }
            else if (isKeyPressed(KEY_NSPIRE_RIGHT) && (*((uint32_t *)0x900F0020) < 0x95)) {
                // Increase contrast.
                *(uint32_t *)0x900F0020 += 1;
            }
            else if (isKeyPressed(KEY_NSPIRE_LEFT) && (*((uint32_t *)0x900F0020) > 0x6B)) {
                // Decrease contrast.
                *(uint32_t *)0x900F0020 -= 1;
            }
        }

        // Cleanup and bye =8-)
        // Restore contrast
        *((uint32_t *)0x900F0020) = initial_contrast;
        free(backbuffer[0]);
    }

    return 0;
}
